@@ -1,3 +1,5 @@
+"""Utilities for smart contract construction."""
+
 import base64
 from typing import Dict, List, NamedTuple, Type, Union
 
@@ -11,9 +13,11 @@ from algosdk.future.transaction import (
 )
 from algosdk.v2client.algod import AlgodClient
 
-Key = Union[str, bytes]
-TType = Union[Type[tl.Int], Type[tl.Bytes]]
-TValue = Union[tl.Int, tl.Bytes]
+from algoappdev.utils import AlgoAppDevError
+
+Key = Union[int, str, bytes]
+TealType = Union[Type[tl.Int], Type[tl.Bytes]]
+TealValue = Union[tl.Int, tl.Bytes]
 
 ZERO = tl.Int(0)
 ONE = tl.Int(1)
@@ -60,32 +64,41 @@ class State:
         Information about an app state key and associated value.
         """
 
-        def __init__(self, key: Key, ttype: TType, default: TValue):
+        def __init__(self, key: Key, type: TealType, default: tl.Expr = None):
             key = self.as_bytes(key)
-            if len(key) > 64:
-                raise ValueError(f"key too long: {key}")
-
             # the key as bytes
             self.key = key
-            # the tl bytes expression
-            self.tbytes = tl.Bytes(key)
             # the tl type of the key's value
-            self.ttype = ttype
+            self.type = type
             # the tl expression to populate the initial value (can be None)
             self.default = default
 
         @staticmethod
         def as_bytes(key: Key) -> bytes:
-            if isinstance(key, bytes):
-                return key
-            else:
+            """
+            Convert a key to it's byte representation. Validates that the key
+            length doesn't surpass the Algorand maximum of 64 bytes.
+            """
+            if isinstance(key, int):
+                # At most 64 keys are allowed, so this fits in a byte. The
+                # byte order is arbitrary as this is handled internally, but
+                # use big-endian anyway to be consistent with TEAL conventions.
+                key = key.to_bytes(1, "big")
+            if isinstance(key, str):
                 return key.encode("utf8")
+            elif isinstance(key, bytes):
+                pass
+            else:
+                raise AlgoAppDevError(f"invalid key type: {type(key)}")
+            if len(key) > 64:
+                raise AlgoAppDevError(f"key too long: {key}")
+            return key
 
     def __init__(self, infos: List[KeyInfo]):
         self._key_to_info = {i.key: i for i in infos}
         self._maybe_values: Dict[bytes, tl.MaybeValue] = {}
 
-    def key_to_info(self, key: Key):
+    def key_info(self, key: Key):
         return self._key_to_info[State.KeyInfo.as_bytes(key)]
 
     def key_infos(self) -> List[KeyInfo]:
@@ -97,9 +110,9 @@ class State:
         num_byte_slices = 0
 
         for info in self._key_to_info.values():
-            if info.ttype is tl.Int:
+            if info.type is tl.Int:
                 num_uints += 1
-            elif info.ttype is tl.Bytes:
+            elif info.type is tl.Bytes:
                 num_byte_slices += 1
 
         return StateSchema(num_uints=num_uints, num_byte_slices=num_byte_slices)
@@ -155,10 +168,10 @@ class StateGlobalExternal(State):
         `MaybeValue` object must be cached if its value is to be accessed more
         than once, so it can remember which slot the value was stored in.
         """
-        info = self.key_to_info(key)
+        info = self.key_info(key)
         maybe_value = self._maybe_values.get(info.key, None)
         if maybe_value is None:
-            maybe_value = tl.App.globalGetEx(self.app_id, info.tbytes)
+            maybe_value = tl.App.globalGetEx(self.app_id, tl.Bytes(info.key))
             self._maybe_values[info.key] = maybe_value
         return maybe_value
 
@@ -209,13 +222,13 @@ class StateGlobal(StateGlobalExternal):
 
     def get(self, key: Key) -> tl.Expr:
         """Build the expression to get the state value at `key`"""
-        info = self.key_to_info(key)
-        return tl.App.globalGet(info.tbytes)
+        info = self.key_info(key)
+        return tl.App.globalGet(tl.Bytes(info.key))
 
-    def set(self, key: Key, value: TValue) -> tl.Expr:
+    def set(self, key: Key, value: TealValue) -> tl.Expr:
         """Build the expression to set the state value at `key`"""
-        info = self.key_to_info(key)
-        return tl.App.globalPut(info.tbytes, value)
+        info = self.key_info(key)
+        return tl.App.globalPut(tl.Bytes(info.key), value)
 
     def constructor(self) -> tl.Expr:
         """
@@ -224,7 +237,7 @@ class StateGlobal(StateGlobalExternal):
         """
         return tl.Seq(
             *[
-                tl.App.globalPut(i.tbytes, i.default)
+                tl.App.globalPut(tl.Bytes(i.key), i.default)
                 for i in self.key_infos()
                 if i.default
             ]
@@ -254,10 +267,12 @@ class StateLocalExternal(State):
 
         See `StateGlobalExternal.get_ex`.
         """
-        info = self.key_to_info(key)
+        info = self.key_info(key)
         maybe_value = self._maybe_values.get(info.key, None)
         if maybe_value is None:
-            maybe_value = tl.App.localGetEx(self.account, self.app_id, info.tbytes)
+            maybe_value = tl.App.localGetEx(
+                self.account, self.app_id, tl.Bytes(info.key)
+            )
             self._maybe_values[info.key] = maybe_value
         return maybe_value
 
@@ -298,13 +313,13 @@ class StateLocal(StateLocalExternal):
 
     def get(self, key: Key) -> tl.Expr:
         """Build the expression to get the state value at `key`"""
-        info = self.key_to_info(key)
-        return tl.App.localGet(self.account, info.tbytes)
+        info = self.key_info(key)
+        return tl.App.localGet(self.account, tl.Bytes(info.key))
 
-    def set(self, key: Key, value: TValue) -> tl.Expr:
+    def set(self, key: Key, value: TealValue) -> tl.Expr:
         """Build the expression to set the state value at `key`"""
-        info = self.key_to_info(key)
-        return tl.App.localPut(self.account, info.tbytes, value)
+        info = self.key_info(key)
+        return tl.App.localPut(self.account, tl.Bytes(info.key), value)
 
     def constructor(self) -> tl.Expr:
         """
@@ -313,7 +328,7 @@ class StateLocal(StateLocalExternal):
         """
         return tl.Seq(
             *[
-                tl.App.localPut(self.account, i.tbytes, i.default)
+                tl.App.localPut(self.account, tl.Bytes(i.key), i.default)
                 for i in self.key_infos()
                 if i.default
             ]
@@ -340,12 +355,12 @@ class AppBuilder(NamedTuple):
       and if it returns zero the app cannot be opted-in by accounts.
     - `on_close_out`: this expression is run when a `CloseOut` transaction is
       sent, and if it returns zero the app cannot be closed out by accounts.
-    - `invokations[name]`: these expressions are run when a `NoOp` transaction
+    - `invocations[name]`: these expressions are run when a `NoOp` transaction
       is sent, and the first argument passed to the call is the bytes encoding
       of `name`.
     - `on_no_op`: this expression is run when a `NoOp` transaction is sent, and
-      no invokation matches the first call argument (if supplied). This is the
-      "default invokation".
+      no invocation matches the first call argument (if supplied). This is the
+      "default invocation".
 
     Branch that executes for a `ClearState` transaction:
 
@@ -361,8 +376,8 @@ class AppBuilder(NamedTuple):
     - opt in is allowed and sets the local state defaults
     - close out is not allowed
     - clear is allowed
-    - no invokations
-    - no default invokation
+    - no invocations
+    - no default invocation
     """
 
     on_create: tl.Expr = None
@@ -371,7 +386,7 @@ class AppBuilder(NamedTuple):
     on_opt_in: tl.Expr = None
     on_close_out: tl.Expr = None
     on_clear: tl.Expr = None
-    invokations: Dict[str, tl.Expr] = None
+    invocations: Dict[str, tl.Expr] = None
     on_no_op: tl.Expr = None
     global_state: StateGlobal = None
     local_state: StateLocal = None
@@ -425,18 +440,18 @@ class AppBuilder(NamedTuple):
                 [tl.Txn.on_completion() == tl.OnComplete.CloseOut, self.on_close_out]
             )
 
-        # handle custon invokations with named arg
-        invokations = {} if self.invokations is None else self.invokations
-        for name, expr in invokations.items():
+        # handle custon invocations with named arg
+        invocations = {} if self.invocations is None else self.invocations
+        for name, expr in invocations.items():
             branches.append(
                 [
-                    # use a an invokation branch for no-op calls with the branch
+                    # use a an invocation branch for no-op calls with the branch
                     # name as arg 0
                     tl.And(
                         tl.Txn.on_completion() == tl.OnComplete.NoOp,
                         tl.If(tl.Txn.application_args.length() >= ONE)
                         # if there is an argument passed, then it must match
-                        # the invokation name
+                        # the invocation name
                         .Then(tl.Txn.application_args[0] == tl.Bytes(name))
                         # otherwise fail the branch
                         .Else(ZERO),
@@ -445,7 +460,7 @@ class AppBuilder(NamedTuple):
                 ]
             )
 
-        # if no invokation matched, then try the default no-op
+        # if no invocation matched, then try the default no-op
         if self.on_no_op:
             branches.append(
                 [tl.Txn.on_completion() == tl.OnComplete.NoOp, self.on_no_op]
